==> text-tree-1.1.0-sources.jar: org/barfuin/texttree/internal/TextTreeImpl.java <==
/*
 * Copyright 2020 the text-tree contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.barfuin.texttree.internal;

import java.util.Iterator;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;

import org.barfuin.texttree.api.Node;
import org.barfuin.texttree.api.TextTree;
import org.barfuin.texttree.api.TreeOptions;
import org.barfuin.texttree.api.style.AnnotationPosition;
import org.barfuin.texttree.api.style.TreeStyle;


/**
 * Renders a text tree.
 *
 * <p>This class is not thread-safe. Use a distinct instance in each thread, if you do multi-threaded tree generation
 * at all 😉.</p>
 */
@NotThreadSafe
public class TextTreeImpl
    implements TextTree
{
    private static final String SPACES = "                                                                          "
        + "                                                                                                         ";

    private final TreeOptions options;

    private ColorizableAppender appender = null;

    private CycleDetector cycleDetector = null;

    /**
     * the width in characters of the printed tree, without annotations. Only needed, and thus only filled, when
     * annotation position is {@link AnnotationPosition#Aligned Aligned}.
     */
    private int treeWidth = -1;



    public TextTreeImpl()
    {
        this(null);
    }



    public TextTreeImpl(@Nullable final TreeOptions pOptions)
    {
        options = pOptions != null ? pOptions : new TreeOptions();
    }



    @Nonnull
    @Override
    public String render(@Nullable final Node pNode)
    {
        init(pNode);

        if (pNode != null) {
            render("", 0, pNode);
        }
        else {
            appender.appendText(null, null);
        }
        return appender.finishUp();
    }



    private void init(@Nullable final Node pNode)
    {
        appender = new ColorizableAppender(options);
        cycleDetector = new CycleDetector(options.getCycleProtection(), options.getIdentityScheme());
        if (options.getAnnotationPosition() == AnnotationPosition.Aligned) {
            treeWidth = precalcTreeWidth(pNode);
        }
    }



    @SuppressWarnings("ConstantConditions")
    private void render(@Nonnull final String pPrefix, final int pLevel, @Nullable final Node pNode)
    {
        final TreeStyle style = options.getStyle();
        printNodeText(pPrefix, pNode);

        final Callout callout = cycleDetector.visit(pNode);
        if (callout != Callout.None) {
            printCallout(callout);
            return;
        }

        if (options.getAnnotationPosition() == AnnotationPosition.NextLine
            || options.getAnnotationPosition() == AnnotationPosition.None
            || pNode == null || pNode.getAnnotation() == null)
        {
            appender.newLine();
        }
        if (pNode != null) {
            printAnnotation(pPrefix, pNode);

            if (hasChildren(pNode)) {
                for (Iterator<? extends Node> iter = pNode.getChildren().iterator(); iter.hasNext(); ) {
                    final boolean maxDepthExceeded = options.getMaxDepth() > 0 && pLevel >= options.getMaxDepth();
                    final Node child = iter.next();

                    appender.append(TreeElementType.Edge, pPrefix);
                    appender.append(TreeElementType.Edge,
                        iter.hasNext() && !maxDepthExceeded ? style.getJunction() : style.getLastJunction());
                    appender.append(TreeElementType.Edge, options.getPaddingStr());

                    if (maxDepthExceeded) {
                        printCallout(Callout.MaxDepth);
                        break;
                    }
                    else {
                        render(pPrefix
                                + (iter.hasNext() ? style.getIndent() : style.getBlankIndent())
                                + options.getPaddingStr(),
                            pLevel + 1, child);
                    }
                }
            }
        }
        cycleDetector.pop();
    }



    private void printNodeText(@Nonnull final String pPrefix, @Nullable final Node pNode)
    {
        final String text = pNode != null ? pNode.getText() : null;
        if (text != null) {
            final String[] lines = text.split(System.lineSeparator());
            for (int i = 0; i < lines.length; i++) {
                final String line = lines[i];
                if (i > 0) {
                    appender.append(TreeElementType.Edge, pPrefix);
                }
                appender.appendText(line, pNode.getColor());
                if (i < lines.length - 1) {
                    appender.newLine();
                }
            }
        }
        else {
            appender.appendText(null, pNode != null ? pNode.getColor() : null);
        }
    }



    private void printCallout(@Nonnull final Callout pCallout)
    {
        final TreeStyle style = options.getStyle();
        if (pCallout.isPrintOnSameLine()) {
            appender.appendText(" ", null);
        }
        appender.append(pCallout.getElementType(), style.getCalloutStart());
        appender.append(pCallout.getElementType(), pCallout.getText());
        appender.append(pCallout.getElementType(), style.getCalloutEnd());
        appender.newLine();
    }



    private void printAnnotation(@Nonnull final String pPrefix, @Nonnull final Node pNode)
    {
        final String annotation = pNode.getAnnotation();
        if (annotation != null && !annotation.isEmpty() && options.getAnnotationPosition() != AnnotationPosition.None) {
            final String[] lines = annotation.trim().split(System.lineSeparator());
            for (int i = 0; i < lines.length; i++) {
                final String line = lines[i];

                if (options.getAnnotationPosition() == AnnotationPosition.NextLine || i > 0) {
                    appender.append(TreeElementType.Edge, pPrefix);
                }

                if (options.getAnnotationPosition() == AnnotationPosition.Inline) {
                    if (i > 0) {
                        appender.append(TreeElementType.Annotation, "  ");  // continuation indent
                    }
                    else {
                        appender.append(TreeElementType.Text, " ");         // separate from node text on same line
                    }
                }

                else if (options.getAnnotationPosition() == AnnotationPosition.Aligned) {
                    int numSpaces = Math.max(1, treeWidth - appender.getCurrentLineLength() + 1);
                    if (i > 0 && hasChildren(pNode)) {
                        appender.append(TreeElementType.Edge, options.getStyle().getIndent());
                        numSpaces -= options.getStyle().getIndent().length();
                    }
                    appender.append(TreeElementType.Annotation, nSpaces(numSpaces));
                }

                appender.append(TreeElementType.Annotation, line);
                appender.newLine();
            }
        }
    }



    @SuppressWarnings("ConstantConditions")
    static boolean hasChildren(@Nullable final Node pNode)
    {
        boolean result = false;
        if (pNode != null && pNode.getChildren() != null) {
            Iterator<? extends Node> iter = pNode.getChildren().iterator();
            return iter != null && iter.hasNext();
        }
        return result;
    }



    private int precalcTreeWidth(@Nullable final Node pNode)
    {
        TreeOptions noAnnotations = TreeOptions.copyOf(options);
        noAnnotations.setAnnotationPosition(AnnotationPosition.None);
        noAnnotations.setColorScheme(null);

        TextTreeImpl textTree = new TextTreeImpl(noAnnotations);
        String bareTree = textTree.render(pNode);
        return longestLine(bareTree);
    }



    int longestLine(@Nonnull final String pMultilineString)
    {
        int result = -1;
        int startOfLine = 0;

        for (int nlPos = pMultilineString.indexOf(System.lineSeparator()); nlPos >= 0;
            nlPos = pMultilineString.indexOf(System.lineSeparator(), startOfLine)) //
        {
            int lineLength = nlPos - startOfLine;
            if (lineLength > result) {
                result = lineLength;
            }
            startOfLine = nlPos + System.lineSeparator().length();
        }

        int lineLength = pMultilineString.length() - startOfLine;
        if (lineLength > result) {
            result = lineLength;
        }
        return result;
    }



    String nSpaces(final int pNumSpaces)
    {
        if (pNumSpaces <= SPACES.length()) {
            return SPACES.substring(0, pNumSpaces);
        }
        return Stream.generate(() -> " ").limit(pNumSpaces).collect(Collectors.joining());
    }
}

==> text-tree-1.2.0-sources.jar: org/barfuin/texttree/internal/TextTreeImpl.java <==
/*
 * Copyright 2020 the text-tree contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.barfuin.texttree.internal;

import java.util.Iterator;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;

import org.barfuin.texttree.api.Node;
import org.barfuin.texttree.api.TextTree;
import org.barfuin.texttree.api.TreeOptions;
import org.barfuin.texttree.api.style.AnnotationPosition;
import org.barfuin.texttree.api.style.TreeStyle;


/**
 * Renders a text tree.
 *
 * <p>This class is not thread-safe. Use a distinct instance in each thread, if you do multi-threaded tree generation
 * at all 😉.</p>
 */
@NotThreadSafe
public class TextTreeImpl
    implements TextTree
{
    private static final String SPACES = "                                                                          "
        + "                                                                                                         ";

    private final TreeOptions options;

    private ColorizableAppender appender = null;

    private CycleDetector cycleDetector = null;

    /**
     * the width in characters of the printed tree, without annotations. Only needed, and thus only filled, when
     * annotation position is {@link AnnotationPosition#Aligned Aligned}.
     */
    private int treeWidth = -1;



    public TextTreeImpl()
    {
        this(null);
    }



    public TextTreeImpl(@Nullable final TreeOptions pOptions)
    {
        options = pOptions != null ? pOptions : new TreeOptions();
    }



    @Nonnull
    @Override
    public String render(@Nullable final Node pNode)
    {
        init(pNode);

        if (pNode != null) {
            render("", 0, pNode, false);
        }
        else {
            appender.appendText(null, null);
        }
        return appender.finishUp();
    }



    private void init(@Nullable final Node pNode)
    {
        appender = new ColorizableAppender(options);
        cycleDetector = new CycleDetector(options);
        if (options.getAnnotationPosition() == AnnotationPosition.Aligned) {
            treeWidth = precalcTreeWidth(pNode);
        }
    }



    @SuppressWarnings("ConstantConditions")
    private void render(@Nonnull final String pPrefix, final int pLevel, @Nullable final Node pNode,
        final boolean pPrecalcWidth)
    {
        final TreeStyle style = options.getStyle();
        printNodeText(pPrefix, pNode);

        final Callout callout = cycleDetector.visit(pNode);
        if (callout != Callout.None) {
            printCallout(callout, pPrecalcWidth);
            cycleDetector.pop();
            return;
        }

        if (options.getAnnotationPosition() == AnnotationPosition.NextLine
            || options.getAnnotationPosition() == AnnotationPosition.None
            || pNode == null || pNode.getAnnotation() == null)
        {
            appender.newLine();
        }
        if (pNode != null) {
            printAnnotation(pPrefix, pNode);

            if (hasChildren(pNode)) {
                for (Iterator<? extends Node> iter = pNode.getChildren().iterator(); iter.hasNext(); ) {
                    final boolean maxDepthExceeded = options.getMaxDepth() > 0 && pLevel >= options.getMaxDepth();
                    final Node child = iter.next();

                    appender.append(TreeElementType.Edge, pPrefix);
                    appender.append(TreeElementType.Edge,
                        iter.hasNext() && !maxDepthExceeded ? style.getJunction() : style.getLastJunction());
                    appender.append(TreeElementType.Edge, options.getPaddingStr());

                    if (maxDepthExceeded) {
                        printCallout(Callout.MaxDepth, pPrecalcWidth);
                        break;
                    }
                    else {
                        render(pPrefix
                                + (iter.hasNext() ? style.getIndent() : style.getBlankIndent())
                                + options.getPaddingStr(),
                            pLevel + 1, child, pPrecalcWidth);
                    }
                }
            }
        }
        cycleDetector.pop();
    }



    private void printNodeText(@Nonnull final String pPrefix, @Nullable final Node pNode)
    {
        final String text = pNode != null ? pNode.getText() : null;
        if (text != null) {
            final String[] lines = text.split(System.lineSeparator());
            for (int i = 0; i < lines.length; i++) {
                final String line = lines[i];
                if (i > 0) {
                    appender.append(TreeElementType.Edge, pPrefix);
                }
                appender.appendText(line, pNode.getColor());
                if (i < lines.length - 1) {
                    appender.newLine();
                }
            }
        }
        else {
            appender.appendText(null, pNode != null ? pNode.getColor() : null);
        }
    }



    private void printCallout(@Nonnull final Callout pCallout, final boolean pPrecalcWidth)
    {
        if (!pPrecalcWidth) {
            final TreeStyle style = options.getStyle();
            if (pCallout.isPrintOnSameLine()) {
                appender.appendText(" ", null);
            }
            appender.append(pCallout.getElementType(), style.getCalloutStart());
            appender.append(pCallout.getElementType(), pCallout.getText());
            appender.append(pCallout.getElementType(), style.getCalloutEnd());
        }
        appender.newLine();
    }



    private void printAnnotation(@Nonnull final String pPrefix, @Nonnull final Node pNode)
    {
        final String annotation = pNode.getAnnotation();
        if (annotation != null && !annotation.isEmpty() && options.getAnnotationPosition() != AnnotationPosition.None) {
            final String[] lines = annotation.trim().split(System.lineSeparator());
            for (int i = 0; i < lines.length; i++) {
                final String line = lines[i];

                if (options.getAnnotationPosition() == AnnotationPosition.NextLine || i > 0) {
                    appender.append(TreeElementType.Edge, pPrefix);
                }

                if (options.getAnnotationPosition() == AnnotationPosition.Inline) {
                    if (i > 0) {
                        appender.append(TreeElementType.Annotation, "  ");  // continuation indent
                    }
                    else {
                        appender.append(TreeElementType.Text, " ");         // separate from node text on same line
                    }
                }

                else if (options.getAnnotationPosition() == AnnotationPosition.Aligned) {
                    int numSpaces = Math.max(1, treeWidth - appender.getCurrentLineLength() + 1);
                    if (i > 0 && hasChildren(pNode)) {
                        appender.append(TreeElementType.Edge, options.getStyle().getIndent());
                        numSpaces -= options.getStyle().getIndent().length();
                    }
                    appender.append(TreeElementType.Annotation, nSpaces(numSpaces));
                }

                appender.append(TreeElementType.Annotation, line);
                appender.newLine();
            }
        }
    }



    @SuppressWarnings("ConstantConditions")
    static boolean hasChildren(@Nullable final Node pNode)
    {
        boolean result = false;
        if (pNode != null && pNode.getChildren() != null) {
            Iterator<? extends Node> iter = pNode.getChildren().iterator();
            return iter != null && iter.hasNext();
        }
        return result;
    }



    private int precalcTreeWidth(@Nullable final Node pNode)
    {
        TreeOptions noAnnotations = TreeOptions.copyOf(options);
        noAnnotations.setAnnotationPosition(AnnotationPosition.None);
        noAnnotations.setColorScheme(null);

        TextTreeImpl textTree = new TextTreeImpl(noAnnotations);
        textTree.init(pNode);
        textTree.render("", 0, pNode, true);
        String bareTree = textTree.appender.finishUp();
        return longestLine(bareTree);
    }



    int longestLine(@Nonnull final String pMultilineString)
    {
        int result = -1;
        int startOfLine = 0;

        for (int nlPos = pMultilineString.indexOf(System.lineSeparator()); nlPos >= 0;
            nlPos = pMultilineString.indexOf(System.lineSeparator(), startOfLine)) //
        {
            int lineLength = nlPos - startOfLine;
            if (lineLength > result) {
                result = lineLength;
            }
            startOfLine = nlPos + System.lineSeparator().length();
        }

        int lineLength = pMultilineString.length() - startOfLine;
        if (lineLength > result) {
            result = lineLength;
        }
        return result;
    }



    String nSpaces(final int pNumSpaces)
    {
        if (pNumSpaces <= SPACES.length()) {
            return SPACES.substring(0, pNumSpaces);
        }
        return Stream.generate(() -> " ").limit(pNumSpaces).collect(Collectors.joining());
    }
}
